/*
** 1998-05-29 -	A rename command might be useful. Since writing this also caused me
**		to implement the generic command interface in cmd_generic, it was
**		very useful indeed!
** 1998-09-18 -	Did some pretty massive changes/additions to provide overwrite protection.
** 1999-01-30 -	Bug fix: always called ovw_overwrite_end() in closing, regardless of
**		whether the _begin() function was ever called. This caused nesting errors.
** 1999-02-23 -	Um... That "bug fix" was buggy. I think I fixed it this time, though.
** 1999-03-05 -	Altered to comply with new selection handling (and its changes on the
**		generic command invocation parameters).
*/

#include "gentoo.h"
#include "cmd_delete.h"
#include "errors.h"
#include "dirpane.h"
#include "overwrite.h"

#include "cmd_generic.h"
#include "cmd_rename.h"

#define	CMD_ID	"rename"

/* ----------------------------------------------------------------------------------------- */

typedef struct {
	GtkWidget	*vbox;
	GtkWidget	*label;
	GtkWidget	*entry;
	MainInfo	*min;
	gboolean	ovw_open;
} RenInfo;

/* ----------------------------------------------------------------------------------------- */

static void ren_body(MainInfo *min, DirPane *src, DirRow2 *row, gpointer gen, gpointer user)
{
	gchar		temp[PATH_MAX + 256];
	const gchar	*name;
	RenInfo		*ren = user;

	name = dp_row_get_name_display(dp_get_tree_model(src), row);
	g_snprintf(temp, sizeof temp, _("Enter New Name For \"%s\""), name);
	gtk_label_set_text(GTK_LABEL(ren->label), temp);
	gtk_entry_set_text(GTK_ENTRY(ren->entry), name);
	gtk_editable_select_region(GTK_EDITABLE(ren->entry), 0, strlen(name));
	gtk_widget_grab_focus(ren->entry);

	cmd_generic_track_entry(gen, ren->entry);

	if(!ren->ovw_open)
	{
		ovw_overwrite_begin(ren->min, _("\"%s\" Already Exists - Proceed With Rename?"), 0U);
		ren->ovw_open = TRUE;
	}
}

static gint ren_action(MainInfo *min, DirPane *src, DirPane *dst, DirRow2 *row, GError **error, gpointer user)
{
	RenInfo		*ren = user;
	const gchar	*old_name, *new_name;
	GFile		*file, *df;
	gboolean	ok;

	old_name = dp_row_get_name(dp_get_tree_model(src), row);
	new_name = gtk_entry_get_text(GTK_ENTRY(ren->entry));

	/* Ignore renames to self. */
	if(strcmp(old_name, new_name) == 0)
	{
		dp_unselect(src, row);
		return 1;
	}

	err_clear(min);

	/* Check for overwrite, and attempt to clear the way. */
	switch(ovw_overwrite_unary_name(src, new_name))
	{
	case OVW_SKIP:
		return 1;
	case OVW_CANCEL:
		return 0;
	case OVW_PROCEED:
		break;
	case OVW_PROCEED_FILE:
	case OVW_PROCEED_DIR:
		df = dp_get_file_from_name(src, new_name);
		if(df == NULL)
			return 0;
		ok = del_delete_gfile(min, df, FALSE, error);
		g_object_unref(df);
		if(!ok)
			return 0;
		break;
	}

	/* All is well, open the file and do the rename. */
	if((file = dp_get_file_from_row(src, row)) != NULL)
	{
		GFile	*nf;

		nf = g_file_set_display_name(file, new_name, NULL, error);
		if(nf != NULL)
			g_object_unref(nf);
		g_object_unref(file);

		return nf != NULL;
	}
	return 0;	/* Fail. */
}

static void ren_free(gpointer user)
{
	if(((RenInfo *) user)->ovw_open)
		ovw_overwrite_end(((RenInfo *) user)->min);
}

/* ----------------------------------------------------------------------------------------- */

gint cmd_rename(MainInfo *min, DirPane *src, DirPane *dst, const CmdArg *ca)
{
	static RenInfo	ri;

	ri.ovw_open = FALSE;
	ri.vbox	    = gtk_vbox_new(FALSE, 0);
	ri.label    = gtk_label_new("");
	ri.entry    = gtk_entry_new();
	gtk_entry_set_max_length(GTK_ENTRY(ri.entry), MAXNAMLEN - 1);
	gtk_box_pack_start(GTK_BOX(ri.vbox), ri.label, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(ri.vbox), ri.entry, FALSE, FALSE, 0);

	return cmd_generic(min, _("Rename"), CGF_NOALL | CGF_NODST, ren_body, ren_action, ren_free, &ri);
}
