/*
** 1998-07-05 -	The Split command, handy for, er, splitting files into parts. Mucho GUI.
** 1999-03-07 -	Altered for the new selection/generic handler. Hum, I really should finish
**		this command off some day...
** 1999-06-19 -	Adapted for the new dialog module. Sheesh, I really should finish this command
**		some day!
*/

#include "gentoo.h"

#include <fcntl.h>

#include "cmd_delete.h"
#include "dialog.h"
#include "dirpane.h"
#include "errors.h"
#include "fileutil.h"
#include "guiutil.h"
#include "overwrite.h"
#include "sizeutil.h"

#include "cmd_generic.h"
#include "cmd_split.h"

#define	CMD_ID	"split"

#define	NFORMAT_LIMIT	(MAXNAMLEN + 16)
#define	SPLIT_CHUNK	(1 << 18)

/* ----------------------------------------------------------------------------------------- */

typedef struct {
	GtkWidget	*vbox;		/* This really is required. */
	GtkWidget	*label;		/* Tells the user what's up. */
	GtkWidget	*mode;		/* Option menu for selecting mode of split. */
	GtkWidget	*mvbox;		/* Sub-vbox for the mode-specific widgets. */
	GtkWidget	*nhbox;		/* Hbox for the naming issue. */
	GtkWidget	*nformat;	/* Entry widget for name formatter. */
	GtkObject	*nbadj;		/* Adjustment for index base. */
	GtkWidget	*nbase;		/* Spin widget for index base. */
	GtkObject	*nsadj;		/* Adjustment for index step. */
	GtkWidget	*nstep;		/* Spin widget for index step. */

	GtkWidget	*nbook;		/* Notebook with mode-settings. */

	GtkWidget	*sssize;	/* Combo holding the wanted size. */
	GtkWidget	*snframe;	/* Frame for "split to fixed number". */

	GtkWidget	*fccount;	/* Spinner for the count, 2+. */

	MainInfo	*min;
	guint		curr_mode;	/* 0 for fixed size, 1 for fixed amount. */
	const gchar	*name;		/* Source file being split. */
} SplitInfo;

/* ----------------------------------------------------------------------------------------- */

/* 1998-09-01 -	Pretty much rewrote this function, now that the sizeutil module exists. */
static void spt_body(MainInfo *min, DirPane *src, DirRow2 *row, gpointer gen, gpointer user)
{
	gchar		buf[2 * MAXNAMLEN], siz1[32], siz2[32];
	SplitInfo	*spi = user;
	GtkTreeModel	*m;
	gint		pos;

	m = dp_get_tree_model(src);
	sze_put_offset(siz1, sizeof siz1, dp_row_get_size(m, row), SZE_BYTES, 3, ',');
	sze_put_offset(siz2, sizeof siz2, dp_row_get_size(m, row), SZE_AUTO, 3, ',');
	if(strcmp(siz1, siz2) != 0)
		g_snprintf(buf, sizeof buf, _("Split \"%s\".\nFile is %s (%s)."), dp_row_get_name_display(m, row), siz1, siz2);
	else
		g_snprintf(buf, sizeof buf, _("Split \"%s\".\nFile is %s."), dp_row_get_name_display(m, row), siz1);
	gtk_label_set_text(GTK_LABEL(spi->label), buf);
	gtk_entry_set_text(GTK_ENTRY(spi->nformat), dp_row_get_name_display(m, row));
	pos = gtk_entry_get_text_length(GTK_ENTRY(spi->nformat));
	gtk_editable_insert_text(GTK_EDITABLE(spi->nformat), ".%03d", 5, &pos);
	spi->name = dp_row_get_name(m, row);

	cmd_generic_track_entry(gen, spi->nformat);
	gtk_combo_box_set_active(GTK_COMBO_BOX(spi->mode), spi->curr_mode);
}

/* 1998-09-01 -	Construct destination file name for piece number <piece> in a split. */
static void build_name(SplitInfo *spi, gchar *buf, gint piece)
{
	gint	base = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(spi->nbase)),
		step = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(spi->nstep)),
		num  = base + step * piece;

	g_snprintf(buf, PATH_MAX, gtk_entry_get_text(GTK_ENTRY(spi->nformat)), num);
}

/* 2009-10-07 -	Do a partial "splice", i.e. a copying of data from an input stream into an output stream.
**		Here, 'chunk' is the number of bytes to copy, and 'buf' & 'buf_size' describe the buffer
**		used to hold the data.
*/
static gboolean streams_splice_partial(GOutputStream *out, GInputStream *in, gsize chunk, gpointer buf, gsize buf_size, GError **err)
{
	gsize	to_go = chunk, bite;
	gssize	did;

	while(to_go > 0)
	{
		bite = to_go > buf_size ? buf_size : to_go;
		did = g_input_stream_read(in, buf, bite, NULL, err);
		if(did == bite)
		{
			did = g_output_stream_write(out, buf, bite, NULL, err);
			if(did != bite)
				break;
		}
		else
			break;
		to_go -= bite;
	}
	return to_go == 0;
}

/* 1998-09-01 -	Split file on row <row> of <src>, producing a bunch of segment files in <dst>.
**		Each segment (except possibly the last) will have the same size.
** 1998-09-12 -	Didn't have any closing of the output files. I really should be shot. Also
**		fixed the protection bits for the output; now same as the source file.
** 1999-11-13 -	Adapted to use new fut_copy_partial() function, fixed bug with closing output
**		file on failed write.
** 2010-07-22 -	Now supports both split modes, in an attack of epic closure.
*/
static gint do_split(SplitInfo *spi, DirPane *src, DirPane *dst, DirRow2 *row, GError **error)
{
	gchar			outname[PATH_MAX];
	goffset			seg_size;
	gsize			to_go = 0, chunk;
	gint			piece = 0;
	GFile			*fin, *fout;
	GFileInputStream	*sin;
	gpointer		tmp;
	gboolean		ok = TRUE;

	if(spi->curr_mode == 0)
	{
		if((seg_size = sze_get_offset(gtk_combo_box_get_active_text(GTK_COMBO_BOX(spi->sssize)))) == (gsize) -1)
			return 0;
	}

	if((fin = dp_get_file_from_row(src, row)) == NULL)
		return 0;
	if((sin = g_file_read(fin, NULL, error)) == NULL)
	{
		g_object_unref(fin);
		return 0;
	}
	if((tmp = g_malloc(SPLIT_CHUNK)) == NULL)
	{
		g_object_unref(fin);
		g_object_unref(sin);
		return 0;
	}

	ovw_overwrite_begin(spi->min, _("\"%s\" Already Exists - Continue With Split?"), 0UL);
	to_go = dp_row_get_size(dp_get_tree_model(src), row);
	if(spi->curr_mode == 1)
	{
		const gint	count = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(spi->fccount));

		seg_size = (gint) (((gfloat) to_go / count) + 0.5f);
	}

	for(; ok && to_go != 0; to_go -= chunk)
	{
		build_name(spi, outname, piece++);
		if((fout = g_file_get_child_for_display_name(dst->dir.root, outname, error)) != NULL)
		{
			GFileOutputStream	*sout;
			OvwRes			ores;

			ores  = ovw_overwrite_unary_file(dst, fout);
			if(ores == OVW_CANCEL)
				ok = FALSE;
			else if(ores == OVW_SKIP)
				ok = g_seekable_seek(G_SEEKABLE(sin), chunk, G_SEEK_CUR, NULL, error);
			else if(ores == OVW_PROCEED_FILE || ores == OVW_PROCEED_DIR)
				ok = del_delete_gfile(spi->min, fout, FALSE, error);
			if(ok && (sout = g_file_create(fout, G_FILE_CREATE_NONE, NULL, error)) != NULL)
			{
				chunk = (to_go > seg_size) ? seg_size : to_go;
				ok = streams_splice_partial(G_OUTPUT_STREAM(sout), G_INPUT_STREAM(sin), chunk, tmp, SPLIT_CHUNK, error);
				g_object_unref(sout);
			}
			else
				ok = FALSE;
			g_object_unref(fout);
		}
		else
			ok = FALSE;
	}
	ovw_overwrite_end(spi->min);
	g_free(tmp);
	g_object_unref(sin);
	g_object_unref(fin);

	return ok;
}

static gint spt_action(MainInfo *min, DirPane *src, DirPane *dst, DirRow2 *row, GError **error, gpointer user)
{
	SplitInfo	*spi = user;
	gint		ret = 0;

	if((ret = do_split(spi, src, dst, row, error)))
		dp_unselect(src, row);

	return ret;
}

/* ----------------------------------------------------------------------------------------- */

/* 1998-07-09 -	Build widgetry needed to support splitting to fixed part size. */
static GtkWidget * build_ss(SplitInfo *spi)
{
	GtkWidget	*frame, *hbox, *label;

	frame = gtk_frame_new(_("Fixed Size Split"));
	hbox  = gtk_hbox_new(FALSE, 0);
	label = gtk_label_new(_("Segment Size"));
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
	spi->sssize = gtk_combo_box_entry_new_text();
	gtk_combo_box_append_text(GTK_COMBO_BOX(spi->sssize), _("1457000 bytes (3.5\" floppy)"));
	gtk_combo_box_append_text(GTK_COMBO_BOX(spi->sssize), _("10485760 bytes (10 MB)"));
	gtk_combo_box_append_text(GTK_COMBO_BOX(spi->sssize), _("26214400 bytes (25 MB)"));
	gtk_combo_box_append_text(GTK_COMBO_BOX(spi->sssize), _("52428800 bytes (50 MB)"));
	gtk_combo_box_append_text(GTK_COMBO_BOX(spi->sssize), _("78643200 bytes (75 MB)"));
	gtk_combo_box_append_text(GTK_COMBO_BOX(spi->sssize), _("100431360 bytes (95 MB, Zip disk)"));
	gtk_box_pack_start(GTK_BOX(hbox), spi->sssize, TRUE, TRUE, 0);

	gtk_container_add(GTK_CONTAINER(frame), hbox);

	return frame;
}

/* 2010-07-22 -	Quite a while later, I actually got around to implementing the other split
**		mode: fixed count (=number of parts).
*/
static GtkWidget * build_fc(SplitInfo *spi)
{
	GtkWidget	*frame, *hbox, *label;

	frame = gtk_frame_new(_("Fixed Count Split"));

	hbox = gtk_hbox_new(FALSE, 0);
	label = gtk_label_new(_("Segment Count"));
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
	spi->fccount = gtk_spin_button_new_with_range(2, 1000, 1);	/* Incredibly arbitrary, yes. */
	gtk_box_pack_start(GTK_BOX(hbox), spi->fccount, TRUE, TRUE, 0);

	gtk_container_add(GTK_CONTAINER(frame), hbox);

	return frame;
}

/* ----------------------------------------------------------------------------------------- */

/* 1998-09-01 -	Callback for mode option menu. */
static void evt_mode_select(GtkWidget *wid, guint index, gpointer user)
{
	SplitInfo	*spi = user;

	spi->curr_mode = index;
	gtk_notebook_set_current_page(GTK_NOTEBOOK(spi->nbook), spi->curr_mode);
}

/* ----------------------------------------------------------------------------------------- */

gint cmd_split(MainInfo *min, DirPane *src, DirPane *dst, const CmdArg *ca)
{
	static SplitInfo	spi = { 0 };
	const gchar		*mode[] = { N_("Fixed size, variable number of parts"),
					    N_("Fixed number of parts, variable sizes"),
					    NULL };
	GtkWidget	*hbox, *label;
	GtkWidget	*hsep;

	spi.min = min;
	spi.vbox = gtk_vbox_new(FALSE, 0);
	spi.label = gtk_label_new(_("Split"));

	/* Build this early, since the signal handler references it. */
	spi.nbook = gtk_notebook_new();
	gtk_notebook_set_show_tabs(GTK_NOTEBOOK(spi.nbook), FALSE);
	gtk_notebook_set_show_border(GTK_NOTEBOOK(spi.nbook), FALSE);
	gtk_notebook_append_page(GTK_NOTEBOOK(spi.nbook), build_ss(&spi), NULL);
	gtk_notebook_append_page(GTK_NOTEBOOK(spi.nbook), build_fc(&spi), NULL);

	hbox = gtk_hbox_new(FALSE, 0);
	label = gtk_label_new(_("Mode"));
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(spi.vbox), spi.label, FALSE, FALSE, 0);
	hsep = gtk_hseparator_new();
	gtk_box_pack_start(GTK_BOX(spi.vbox), hsep, FALSE, FALSE, 4);
	spi.mode = gui_build_combo_box(mode, G_CALLBACK(evt_mode_select), &spi);
	gtk_box_pack_start(GTK_BOX(hbox), spi.mode, TRUE, TRUE, 0);
	gtk_box_pack_start(GTK_BOX(spi.vbox), hbox, FALSE, FALSE, 0);
	spi.nhbox = gtk_hbox_new(FALSE, 0);
	label = gtk_label_new(_("Name Format"));
	gtk_box_pack_start(GTK_BOX(spi.nhbox), label, FALSE, FALSE, 2);
	spi.nformat = gtk_entry_new();
	gtk_entry_set_max_length(GTK_ENTRY(spi.nformat), NFORMAT_LIMIT);
	gtk_box_pack_start(GTK_BOX(spi.nhbox), spi.nformat, TRUE, TRUE, 0);
	label = gtk_label_new(_("Base"));
	gtk_box_pack_start(GTK_BOX(spi.nhbox), label, FALSE, FALSE, 0);
	spi.nbadj = gtk_adjustment_new(0, 0, 31, 1, 16, 0.);
	spi.nbase = gtk_spin_button_new(GTK_ADJUSTMENT(spi.nbadj), 0, 0);
	gtk_box_pack_start(GTK_BOX(spi.nhbox), spi.nbase, FALSE, FALSE, 0);
	label = gtk_label_new(_("Step"));
	gtk_box_pack_start(GTK_BOX(spi.nhbox), label, FALSE, FALSE, 0);
	spi.nsadj = gtk_adjustment_new(1, 1, 31, 1, 16, 0.);
	spi.nstep = gtk_spin_button_new(GTK_ADJUSTMENT(spi.nsadj), 0, 0);
	gtk_box_pack_start(GTK_BOX(spi.nhbox), spi.nstep, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(spi.vbox), spi.nhbox, FALSE, FALSE, 2);

	gtk_box_pack_start(GTK_BOX(spi.vbox), spi.nbook, TRUE, TRUE, 0);

	return cmd_generic(min, _("Split"), CGF_NOALL | CGF_NODIRS, spt_body, spt_action, NULL, &spi);
}
